from yta_editor_time.specifications.frame import FrameIndexSpecification
from yta_editor_time.specifications.time import TimeSpecification
from yta_editor_time.specifications.progress import ProgressSpecification


__all__ = [
    'FrameIndexSpecification',
    'TimeSpecification',
    'ProgressSpecification'
]


from yta_constants.enum import YTAEnum as Enum


# TODO: By now we are not implementing the 'anchor'
# because it complicates the logic a lot...
# class Anchor(Enum):
#     """
#     The part of the element the settings are anchored
#     to and must be considered when calculating
#     according to the context given.
#     """

#     START = 'start'
#     """
#     Anchored to the `start` of the element and depending
#     on it.
#     """
#     END = 'end'
#     """
#     Anchored to the `end` of the element and depending on
#     it.
#     """

# TODO: Move inside 'specifications'
class SpecificationRequirement(Enum):
    """
    Enum class to define the requirements a
    specification has.
    """

    FPS = 'fps'
    TOTAL_FRAMES = 'total_frames'
    DURATION = 'duration'



from yta_editor_time.specifications.types import FrameIndex
from yta_editor_time.specifications.resolved import FrameIndexSpecificationResolved


class SpecificationEvaluator:
    """
    Class to simplify the way we evaluate if a
    `frame_index` is active or not according to
    a specification that has been previously
    resolved.
    """

    @staticmethod
    def is_active(
        frame_index: FrameIndex,
        frame_index_specification_resolved: FrameIndexSpecificationResolved
    ) -> bool:
        """
        Check if the `frame_index` provided is active for
        the `frame_index_specification_resolved` provided.
        """
        return int(frame_index_specification_resolved.start_frame) <= int(frame_index) < int(frame_index_specification_resolved.end_frame)


from yta_editor_time.enum import SpecificationRequirement
from yta_editor_time.specifications.types import TimeMoment, TimeDuration
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union

    
@dataclass(frozen = True)
class TimeSpecification:
    """
    The time range of the element that must be affected
    according to this specification.

    TODO: Explain like the others
    """

    start_t: Union[TimeMoment, None]
    """
    The first time moment that must be affected by this
    specification.
    """
    end_t: Union[TimeMoment, None]
    """
    The last time moment that must be affected by this
    specification.
    """
    duration: Union[TimeDuration, None] = None
    """
    The total duration that this specification must
    last.
    """

    @property
    def requirements(
        self
    ) -> dict[SpecificationRequirement]:
        """
        The requirements of this specification that are
        needed to be able to resolve it.
        """
        return {
            SpecificationRequirement.FPS
        }

    def __post_init__(
        self
    ):
        if (
            not PythonValidator.is_instance_of(self.start_t, TimeMoment) and
            self.start_t is not None
        ):
            object.__setattr__(self, 'start_t', TimeMoment(self.start_t))
        if (
            not PythonValidator.is_instance_of(self.end_t, TimeMoment) and
            self.end_t is not None
        ):
            object.__setattr__(self, 'end_t', TimeMoment(self.end_t))
        if (
            not PythonValidator.is_instance_of(self.duration, TimeDuration) and
            self.duration is not None
        ):
            object.__setattr__(self, 'duration', TimeDuration(self.duration))

        # exactly two of three must be provided
        provided = sum(
            x is not None
            for x in (self.start_t, self.end_t, self.duration)
        )

        if provided != 2:
            raise ValueError(
                'Provide exactly two of: start_t, end_t, duration'
            )

        if (
            self.start_t and
            self.end_t
        ):
            if self.end_t.value < self.start_t.value:
                raise ValueError('end_t must be >= start_t')

        if (
            self.duration and
            self.duration.value == 0
        ):
            raise ValueError('duration must be > 0')


from yta_editor_time.enum import SpecificationRequirement
from yta_editor_time.specifications.types import FrameIndex, FramesNumber
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union


@dataclass(frozen = True)
class FrameIndexSpecification:
    """
    The frames that must be affected according to
    this specification.

    Here you have some instructions about the
    different options you have:
    - Set `start_frame` and `end_frame` to apply
    the modification to the frames from the
    `start_frame` to the `end_frame` (including not
    this `end_frame`).
    - Set `start_frame` and `total_frames` to apply
    the modification from the `start_frame` and 
    lasting the `total_frames` provided.
    """

    start_frame: Union[FrameIndex, None]
    """
    The first frame that must be affected by this
    specification.
    """
    end_frame: Union[FrameIndex, None]
    """
    The last frame that must be affected by this
    specification.
    """
    total_frames: Union[FramesNumber, None]
    """
    The total amount of frames that this specification
    must last.
    """

    @property
    def requirements(
        self
    ) -> dict[SpecificationRequirement]:
        """
        The requirements of this specification that are
        needed to be able to resolve it.
        """
        return {
            SpecificationRequirement.TOTAL_FRAMES
        }

    def __post_init__(
        self
    ):
        if (
            not PythonValidator.is_instance_of(self.start_frame, FrameIndex) and
            self.start_frame is not None
        ):
            object.__setattr__(self, 'start_frame', FrameIndex(self.start_frame))
        if (
            not PythonValidator.is_instance_of(self.end_frame, FrameIndex) and
            self.end_frame is not None
        ):
            object.__setattr__(self, 'end_frame', FrameIndex(self.end_frame))
        if (
            not PythonValidator.is_instance_of(self.total_frames, FramesNumber) and
            self.total_frames is not None
        ):
            object.__setattr__(self, 'total_frames', FramesNumber(self.total_frames))

        # exactly two of three must be provided
        provided = sum(
            x is not None
            for x in (self.start_frame, self.end_frame, self.total_frames)
        )

        if provided != 2:
            raise ValueError(
                'Provide exactly two of: start_frame, end_frame, total_frames'
            )

        if (
            self.start_frame and
            self.end_frame
        ):
            if self.end_frame.value < self.start_frame.value:
                raise ValueError('end_frame must be >= start_frame')

        if (
            self.total_frames and
            self.total_frames.value == 0
        ):
            raise ValueError('total_frames must be > 0')


from yta_editor_time.specifications.types import FrameIndex
from dataclasses import dataclass


@dataclass(frozen = True)
class FrameIndexSpecificationResolved:
    """
    A frame index specification that has been
    resolved according to a specific evaluation
    context.

    This resolved specification will indicate the
    frames range in which the transformations
    should be applied, from the `start_frame` to
    the `end_frame`.
    """

    start_frame: FrameIndex
    """
    The frame in which the modifications should
    start being applied.
    """
    end_frame: FrameIndex  # exclusive
    """
    The frame in which the modifications should
    stop being applied, including it not.
    """



from yta_editor_time.specifications import FrameIndexSpecification, TimeSpecification, ProgressSpecification
from yta_editor_time.specifications.types import FrameIndex, TimeMoment, ProgressValue, FramesNumber, Fps, TimeDuration
from yta_editor_time.specifications.resolved import FrameIndexSpecificationResolved
from yta_validation import PythonValidator
from typing import Union


class SpecificationResolver:
    """
    *Static class*

    Class to simplify the way we resolve the
    different specifications.
    """

    @staticmethod
    def resolve(
        specification: Union[ProgressSpecification, TimeSpecification, FrameIndexSpecification],
        fps: Union[Fps, None] = None,
        total_frames: Union[FramesNumber, None] = None,
        duration: Union[TimeDuration, None] = None
    ):
        return (
            SpecificationResolver._resolve_progress_specification(
                progress_specification = specification,
                fps = fps,
                duration = duration
            )
            if PythonValidator.is_instance_of(specification, ProgressSpecification) else
            SpecificationResolver._resolve_time_specification(
                time_specification = specification,
                fps = fps
            )
            if PythonValidator.is_instance_of(specification, TimeSpecification) else
            SpecificationResolver._resolve_frame_index_specification(
                frame_index_specification = specification,
                total_frames = total_frames
            )
        )

    @staticmethod
    def _resolve_frame_index_specification(
        frame_index_specification: FrameIndexSpecification,
        total_frames: FramesNumber
    ) -> FrameIndexSpecificationResolved:
        """
        *For internal use only*

        Resolve the `frame_index_specification` provided
        considering the `total_frames` provided.
        """
        if (
            frame_index_specification.start_frame and
            frame_index_specification.end_frame
        ):
            start_frame = frame_index_specification.start_frame
            end_frame = frame_index_specification.end_frame
        elif (
            frame_index_specification.start_frame and
            frame_index_specification.total_frames
        ):
            start_frame = frame_index_specification.start_frame
            end_frame = FrameIndex(start_frame.value + frame_index_specification.total_frames.value)
        elif (
            frame_index_specification.end_frame and
            frame_index_specification.total_frames
        ):
            end_frame = frame_index_specification.end_frame
            start_frame = FrameIndex(end_frame.value - frame_index_specification.total_frames.value)
        else:
            raise AssertionError('Unreachable')

        if (
            start_frame.value < 0 or
            end_frame.value > int(total_frames)
        ):
            raise ValueError('Resolved values out of bounds')

        return FrameIndexSpecificationResolved(
            start_frame = start_frame,
            end_frame = end_frame
        )
    
    def _resolve_time_specification(
        time_specification: TimeSpecification,
        fps: Fps
    ) -> FrameIndexSpecificationResolved:
        """
        *For internal use only*

        Resolve the `time_specification` provided
        considering the `fps` provided.
        """
        def to_frame(
            t: TimeMoment
        ) -> FrameIndex:
            from yta_video_frame_time.t_media import TMedia

            return FrameIndex(
                value = TMedia(
                    t = float(t),
                    fps = float(fps)
                ).frame_index
            )

        if (
            time_specification.start_t and
            time_specification.end_t
        ):
            start_frame = to_frame(time_specification.start_t)
            end_frame = to_frame(time_specification.end_t)
        elif (
            time_specification.start_t and
            time_specification.duration
        ):
            start_frame = to_frame(time_specification.start_t)
            end_frame = to_frame(TimeMoment(time_specification.start_t.value + time_specification.duration.value))
        elif (
            time_specification.end_t and
            time_specification.duration
        ):
            end_frame = to_frame(time_specification.end_t)
            start_frame = to_frame(TimeMoment(time_specification.end_t.value - time_specification.duration.value))
        else:
            raise ValueError

        return FrameIndexSpecificationResolved(
            start_frame = start_frame,
            end_frame = end_frame
        )

    def _resolve_progress_specification(
        progress_specification: ProgressSpecification,
        fps: Fps,
        duration: TimeDuration
    ) -> FrameIndexSpecificationResolved:
        """
        *For internal use only*

        Resolve the `progress_specification` provided
        considering the `fps` and the `duration` provided.
        """
        def to_frame(
            progress: ProgressValue
        ) -> FrameIndex:
            from yta_video_frame_time.t_media import TMedia

            return FrameIndex(
                value = TMedia.from_progress_normalized(
                    progress_normalized = float(progress),
                    duration = float(duration),
                    fps = float(fps)
                ).frame_index
            )

        if (
            progress_specification.start_progress and
            progress_specification.end_progress
        ):
            start_frame = to_frame(progress_specification.start_progress)
            end_frame = to_frame(progress_specification.end_progress)
        elif (
            progress_specification.start_progress and
            progress_specification.duration
        ):
            start_frame = to_frame(progress_specification.start_progress)
            end_frame = to_frame(
                ProgressValue(progress_specification.start_progress.value + progress_specification.duration.value)
            )
        elif (
            progress_specification.end_progress and
            progress_specification.duration
        ):
            end_frame = to_frame(progress_specification.end_progress)
            start_frame = to_frame(
                ProgressValue(progress_specification.end_progress.value - progress_specification.duration.value)
            )
        else:
            raise ValueError

        return FrameIndexSpecificationResolved(
            start_frame = start_frame,
            end_frame = end_frame
        )
    
    


from yta_video_frame_time.t_media import TMedia
from yta_editor_time.specifications.types import FrameIndex, TimeMoment, ProgressValue, FramesNumber, Fps, TimeDuration
from yta_validation import PythonValidator
from quicktions import Fraction
from dataclasses import dataclass, field
from typing import Union


# TODO: Maybe this should be moved to another project (?)
@dataclass(frozen = True)
class EvaluationContext:
    """
    A context to be evaluated in order to make
    calculations to apply effects, transitions, etc.

    Each instance is related to an specific scope
    and don't need information about the rest.
    """

    fps: Fps = Fps(60.0)
    """
    The frames per second.
    """
    total_frames: FramesNumber = FramesNumber(60)
    """
    The total amount of frames, which is also the
    duration but in number of frames.
    """
    frame_index: FrameIndex = FrameIndex(0)
    """
    The frame index, which is the most exact value.

    This value is the only source of truth for the
    whole evaluation context.
    """
    t: TimeMoment = field(init = False)
    """
    The time moment, truncated and multiple of `1/fps`
    that corresponds to the `frame_index`.

    This value is autocalculated in the `__post_init__`
    method.

    The formula:
    - `frame_index * (1 / fps)`
    """
    progress: ProgressValue = field(init = False)
    """
    The progress normalized that corresponds to the
    given `frame_index` according to the `total_frames`
    given.

    This value is autocalculated in the `__post_init__`
    method.

    The formula:
    - `frame_index / (duration_in_frames - 1)`
    """
    duration: TimeDuration = field(init = False)
    """
    The duration as a `Fraction` value.

    This value is autocalculated in the `__post_init__`
    method.

    The formula:
    - `total_frames * (1 / fps)`
    """

    def __post_init__(
        self
    ):
        if not PythonValidator.is_instance_of(self.fps, Fps):
            object.__setattr__(self, 'fps', Fps(self.fps))
        if not PythonValidator.is_instance_of(self.total_frames, FramesNumber):
            object.__setattr__(self, 'total_frames', FramesNumber(self.total_frames))
        if not PythonValidator.is_instance_of(self.frame_index, FrameIndex):
            object.__setattr__(self, 'frame_index', FrameIndex(self.frame_index))
        
        t_media = TMedia.from_frame_index(
            frame_index = int(self.frame_index),
            fps = float(self.fps)
        ).t
        progress_normalized = TMedia.to_progress_normalized(
            t = t_media,
            # TODO: Maybe in frames (?)
            duration = int(self.total_frames) / float(self.fps),
            fps = float(self.fps)
        )
        duration = TMedia.from_frame_index(
            frame_index = int(self.total_frames),
            fps = float(self.fps)
        ).t

        object.__setattr__(self, 't', TimeMoment(t_media))
        object.__setattr__(self, 'progress', ProgressValue(progress_normalized))
        object.__setattr__(self, 'duration', TimeDuration(duration))

    @classmethod
    def from_t(
        cls,
        fps: float,
        total_frames: int,
        t: Union[float, Fraction]
    ) -> 'EvaluationContext':
        """
        Get a new `EvaluationContext` instance by using
        the `t` time moment provided with the `fps` and
        `total_frames` also included.
        """
        return cls(
            fps = fps,
            total_frames = total_frames,
            frame_index = TMedia(
                t = t,
                fps = fps
            ).frame_index
        )


from yta_editor_time.enum import SpecificationRequirement
from yta_editor_time.specifications.types import ProgressValue, TimeDuration
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union

    
@dataclass(frozen = True)
class ProgressSpecification:
    """
    The part of the element that must be affected,
    defined by a progress from its start to its end
    as a normalized value in a `[0.0, 1.0]` range.

    Here you have some instructions about the
    different options you have, considering the
    progress as a value from `0.0` (start of the
    element) to `1.0` (end of the element):
    - Set `start_progress` and `end_progress` to
    apply the modification to that progress range.
    - Set `start_progress` and `duration` to start
    applying the modification at the `start_progress`
    moment and lasting the `duration` seconds
    provided.
    """

    start_progress: Union[ProgressValue, None]
    """
    The start of the progress that must be affected by
    this specification.
    """
    end_progress: Union[ProgressValue, None]
    """
    The end of the progress that must be affected by
    this specification.
    """
    duration: Union[TimeDuration, None] = None
    """
    The total duration that this specification must
    last.
    """

    @property
    def requirements(
        self
    ) -> dict[SpecificationRequirement]:
        """
        The requirements of this specification that are
        needed to be able to resolve it.
        """
        return {
            SpecificationRequirement.FPS,
            SpecificationRequirement.DURATION
        }

    def __post_init__(
        self
    ):
        if (
            not PythonValidator.is_instance_of(self.start_progress, ProgressValue) and
            self.start_progress is not None
        ):
            object.__setattr__(self, 'start_progress', ProgressValue(self.start_progress))
        if (
            not PythonValidator.is_instance_of(self.end_progress, ProgressValue) and
            self.end_progress is not None
        ):
            object.__setattr__(self, 'end_progress', ProgressValue(self.end_progress))
        if (
            not PythonValidator.is_instance_of(self.duration, TimeDuration) and
            self.duration is not None
        ):
            object.__setattr__(self, 'duration', TimeDuration(self.duration))

        # exactly two of three must be provided
        provided = sum(
            x is not None
            for x in (self.start_progress, self.end_progress, self.duration)
        )

        if provided != 2:
            raise ValueError(
                'Provide exactly two of: start_progress, end_t, duration'
            )

        if (
            self.start_progress and
            self.end_progress
        ):
            if self.end_progress.value < self.start_progress.value:
                raise ValueError('end_progress must be >= start_progress')

        if (
            self.duration and
            self.duration.value == 0
        ):
            raise ValueError('duration must be > 0')
        
        # TODO: Maybe raise exception if 'start_progress' +
        # 'duration' > 1.0 (?)